/*
 * @(#)BevelU.c
 *
 * Copyright 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Undo algorithm for Bevel */

#include "BevelP.h"

static BevelLoc startLoc[MAX_FACES][MAX_CUBES];

static void
newStack(BevelStack *s)
{
	if (s->lastMove != NULL || s->firstMove != NULL)
		return;
	if (!(s->lastMove = (MoveStack *) malloc(sizeof (MoveStack)))) {
		DISPLAY_ERROR("Not enough memory, exiting.");
	}
	if (!(s->firstMove = (MoveStack *) malloc(sizeof (MoveStack)))) {
		DISPLAY_ERROR("Not enough memory, exiting.");
	}
	s->firstMove->previous = s->lastMove->next = NULL;
	s->firstMove->next = s->lastMove;
	s->lastMove->previous = s->firstMove;
	s->count = 0;
}

static void
pushStack(BevelStack *s, moveRecord **move)
{
	if (!(s->currMove = (MoveStack *) malloc(sizeof (MoveStack)))) {
		DISPLAY_ERROR("Not enough memory, exiting.");
	}
	s->lastMove->previous->next = s->currMove;
	s->currMove->previous = s->lastMove->previous;
	s->currMove->next = s->lastMove;
	s->lastMove->previous = s->currMove;
	*move = &(s->currMove->move);
	s->count++;
}

static void
popStack(BevelStack *s)
{
	s->currMove = s->lastMove->previous;
	s->lastMove->previous->previous->next = s->lastMove;
	s->lastMove->previous = s->lastMove->previous->previous;
	free(s->currMove);
	s->count--;
}

static moveRecord *
topStack(BevelStack *s)
{
	return &(s->lastMove->previous->move);
}

static int
emptyStack(BevelStack *s)
{
	return (s->lastMove->previous == s->firstMove);
}

static void
flushStack(BevelStack *s)
{
	while (s->lastMove->previous != s->firstMove) {
		s->currMove = s->lastMove->previous;
		s->lastMove->previous->previous->next = s->lastMove;
		s->lastMove->previous = s->lastMove->previous->previous;
		free(s->currMove);
	}
	s->count = 0;
}

static void
deleteStack(BevelStack *s)
{
	flushStack(s);
	if (s->firstMove) {
		free(s->firstMove);
		s->firstMove = NULL;
	}
	if (s->lastMove) {
		free(s->lastMove);
		s->lastMove = NULL;
	}
}

/**********************************/

void
newMoves(BevelStack *s)
{
	newStack(s);
}

void
deleteMoves(BevelStack *s)
{
	deleteStack(s);
}

static void
writeMove(moveRecord *move, int direction, int control,
		int face, int position)
{
#if 0
	move->direction = direction;
	move->control = control;
	move->face = face;
	move->position = position;
#endif
	move->packed = (unsigned short int) (((control & 0xF) << 12) +
		((direction & 0xF) << 8) + ((position & 0xF) << 4) +
		(face & 0xF));
}

static void
readMove(moveRecord *move, int *direction, int *control,
		int *face, int *position)
{
#if 0
	*direction = move->direction;
	*control = move->control;
	*face = move->face;
	*position = move->position;
#endif
	*face = (int) (move->packed & 0xF);
	*position = (int) ((move->packed >> 4) & 0xF);
	*direction = (int) ((move->packed >> 8) & 0xF);
	*control = (int) ((move->packed >> 12) & 0xF);
}

void
setMove(BevelStack *s, int direction, int control,
		int face, int position)
{
	moveRecord *move;

	pushStack(s, &move);
	writeMove(move, direction, control, face, position);
}

void
getMove(BevelStack *s, int *direction, int *control,
		int *face, int *position)
{
	readMove(topStack(s), direction, control, face, position);
	popStack(s);
}

int
madeMoves(BevelStack *s)
{
	return !emptyStack(s);
}

void
flushMoves(BevelWidget w, BevelStack *s, Boolean undo)
{
	int face, position;

	flushStack(s);
	if (undo) {
		for (face = 0; face < MAX_FACES; face++)
			for (position = 0; position < MAX_CUBES; position++) {
				startLoc[face][position].face =
					w->bevel.cubeLoc[face][position].face;
				startLoc[face][position].rotation =
					w->bevel.cubeLoc[face][position].rotation;
			}
	}
}

int
numMoves(BevelStack *s)
{
	return s->count;
}

Boolean
scanMoves(FILE *fp, BevelWidget w, int moves)
{
	int direction, control, face, position, k, c;

	for (k = 0; k < moves; k++) {
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		if (fscanf(fp, "%d %d %d %d",
				&direction, &control, &face, &position) != 4) {
			(void) fprintf(stderr,
				"corrupt scan, expecting 4 integers for move %d\n", k);
			return False;
		}
		movePuzzle(w, face, position, direction, control, INSTANT);
	}
	return True;
}

void
printMoves(FILE *fp, BevelStack *s)
{
	int face, position, direction, control, counter = 0;

	s->currMove = s->firstMove->next;
	(void) fprintf(fp, "moves\tdir\tcontrol\tface\tposition\n");
	while (s->currMove != s->lastMove) {
		readMove(&(s->currMove->move),
			&direction, &control, &face, &position);
		(void) fprintf(fp, "%d%c\t%d\t%d\t%d\t%d\n",
			++counter, SYMBOL, direction, control, face, position);
		s->currMove = s->currMove->next;
	}
}

Boolean
scanStartPosition(FILE *fp, BevelWidget w)
{
	int face, position, num, c;

	while ((c = getc(fp)) != EOF && c != SYMBOL);
	for (face = 0; face < MAX_FACES; face++)
		for (position = 0; position < MAX_CUBES; position++) {
			if (fscanf(fp, "%d ", &num) != 1) {
				(void) fprintf(stderr,
					"corrupt start record, expecting an integer for %d %d\n",
					face, position);
				return False;
			}
			startLoc[face][position].face = num;
			if (w->bevel.orient) {
				if (fscanf(fp, "%d ", &num) != 1) {
					(void) fprintf(stderr,
						"corrupt start orient record, expecting an integer for %d %d\n",
						face, position);
					return False;
				}
				startLoc[face][position].rotation = num;
			}
		}
	return True;
}

void
printStartPosition(FILE *fp, BevelWidget w)
{
	int face, position;

	(void) fprintf(fp, "\nstartingPosition%c\n", SYMBOL);
	for (face = 0; face < MAX_FACES; face++) {
		for (position = 0; position < MAX_CUBES; position++) {
			(void) fprintf(fp, "%4d", startLoc[face][position].face);
			if (w->bevel.orient) {
				(void) fprintf(fp, "%3d",
					startLoc[face][position].rotation);
			}
		}
		(void) fprintf(fp, "\n");
	}
	(void) fprintf(fp, "\n");
}

void
setStartPosition(BevelWidget w)
{
	int face, position;

	for (face = 0; face < MAX_FACES; face++)
		for (position = 0; position < MAX_CUBES; position++) {
			w->bevel.cubeLoc[face][position].face =
				startLoc[face][position].face;
			if (w->bevel.orient)
				w->bevel.cubeLoc[face][position].rotation =
					startLoc[face][position].rotation;
		}
	drawAllPieces(w);
}
